/*
 * cursplash.c
 * curses splash with progress bar for OpenRC
 */

/*
 * Copyright 2008 Roy Marples <roy@marples.name>
 * All rights reserved

 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

const char cs_copyright[] = "Copyright (c) 2008 Roy Marples";

#define _BSD_SOURCE

#include <sys/ioctl.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/termios.h>
#include <sys/types.h>
#include <sys/un.h>

#if defined(__linux__)
# include <linux/vt.h>
#elif defined(__NetBSD__) || defined(__OpenBSD__)
# include <dev/wscons/wsdisplay_usl_io.h>
#endif

#include <curses.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <einfo.h>
#include <rc.h>

#define UNCONST(a)	((void *)(unsigned long)(const void *)(a))

#define LOGO		"/etc/cursplash.logo"
#define SOCKET		"/lib/rc/init.d/cursplash.sock"
#define NFDS		5

#if defined(__linux__)
# define SPLASH_TTY		17
# define SPLASH_TTY_PATH	"/dev/tty17"
#else
# define SPLASH_TTY		5
# define SPLASH_TTY_PATH	"/dev/ttyE4"
#endif

static struct pollfd fds[NFDS];
static struct sockaddr_un sun;

static RC_STRINGLIST *svc_todo;
static RC_STRINGLIST svc_done;
static int starting;

static WINDOW *mainwnd;
static WINDOW *mbox;

static int stdout_fd = -1;
static int stderr_fd = -1;
static int console_fd = -1;
static int splash_fd = -1;
static int orig_vt = -1;

static void
cleanup(void)
{
	unlink(SOCKET);
	erase();
	refresh();
	endwin();
	close(splash_fd);
#ifdef __linux__
	vhangup();
#endif
	if (orig_vt != -1) {
		ioctl(console_fd, VT_ACTIVATE, orig_vt);
		ioctl(console_fd, VT_WAITACTIVE, orig_vt);
	}
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
#ifdef VT_DISALLOCATE
	ioctl(console_fd, VT_DISALLOCATE, SPLASH_TTY);
#endif
	close(console_fd);
}

static void
splash_message(const char *message, const char *service)
{
	char buffer[1024], *p;
	int cols, i;

	cols = getmaxx(mbox) - 2;
	p = buffer;
	memset(buffer, 0, sizeof(buffer));
	p += snprintf(buffer, sizeof(buffer), " %s %s", message, service);
	for (i = (p - buffer); i < cols; i++)
		*p++ = ' ';
	*p = '\0';
	wcolor_set(mbox, COLOR_BLUE, NULL);
	mvwprintw(mbox, 1, 1, "%s", buffer);
	wrefresh(mbox);
	refresh();
}

static void
splash_progress(void)
{
	double done = 0, todo = 0, total, perc;
	int cols, split, i;
	double upto;
	RC_STRING *s;
	char buffer[1024], *p;

	if (svc_todo) {
		TAILQ_FOREACH(s, svc_todo, entries)
			todo++;
		TAILQ_FOREACH(s, &svc_done, entries)
			done++;
	}
	total = todo + done;
	cols = getmaxx(mbox) - 4; 
	split = (cols - 6) / 2;
	upto = (cols / total) * done;
	p = buffer;
	*p++ = ' ';
	for (i = 0; i < cols; i++) {
		if (i == split) {
			if (total)
				perc = (100 * done) / total;
			else
				perc = 0;
			p += snprintf(p, sizeof(buffer) - (p - buffer),
				      " %02d%s%% ", (int)perc,
				      (total && done == total) ? "" : " ");
			i += 5;
		} else if (i < upto)
			*p++ = '#';
		else
			*p++ = ' ';
	}
	*p++ = ' ';
	*p = '\0';
	wattron(mbox, A_BOLD);
	wcolor_set(mbox, COLOR_YELLOW, NULL);
	mvwprintw(mbox, 2, 1, "%s", buffer);
	wrefresh(mbox);
}

static int
service_starting(const char *service)
{
	RC_STRING *s;

	if (!starting)
		return 0;
	TAILQ_FOREACH(s, svc_todo, entries)
		if (s->value && strcmp(s->value, service) == 0)
			break;
	if (s)
		splash_message("Starting", service);
	return 0;
}

static int
service_started(const char *service)
{
	RC_STRING *s;

	if (!starting)
		return 0;
	TAILQ_FOREACH(s, svc_todo, entries)
		if (strcmp(s->value, service) == 0)
			break;
	if (s) {
		TAILQ_REMOVE(svc_todo, s, entries);
		TAILQ_INSERT_TAIL(&svc_done, s, entries);
		splash_progress();
		splash_message("Started", service);
	}
	return 0;
}

static int
service_stopping(const char *service)
{
	RC_STRING *s;

	if (starting)
		return 0;
	TAILQ_FOREACH(s, svc_todo, entries)
		if (strcmp(s->value, service) == 0)
			break;
	if (s)
		splash_message("Stopping", service);
	return 0;
}

static int
service_stopped(const char *service)
{
	RC_STRING *s;

	if (starting)
		return 0;
	TAILQ_FOREACH(s, svc_todo, entries)
		if (strcmp(s->value, service) == 0)
			break;
	if (s) {
		TAILQ_REMOVE(svc_todo, s, entries);
		TAILQ_INSERT_TAIL(&svc_done, s, entries);
		splash_progress();
		splash_message("Stopped", service);
	}
	return 0;
}

#ifdef __linux__
static int
closefrom(int fd)
{
	struct rlimit rl;
	size_t i;
	int r = 0;

	if (getrlimit(RLIMIT_NOFILE, &rl) == -1)
		return -1;
	for (i = fd; i < rl.rlim_cur; i++)
		r += close(i);
	return r;
}
#endif

static int
set_nonblock(int fd)
{
	int flags;

	if ((flags = fcntl(fd, F_GETFL, 0)) == -1
	    || fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
	{
		return -1;
	}
	return 0;
}

static int
make_sock(socklen_t *len)
{
	int fd;

	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		return -1;
	memset(&sun, 0, sizeof(sun));
	sun.sun_family = AF_UNIX;
	snprintf(sun.sun_path, sizeof(sun.sun_path), "%s", SOCKET);
	if (len)
		*len = sizeof(sun.sun_family) + strlen(sun.sun_path) + 1;
	return fd;
}

static int
process_message(int fd)
{
	char buffer[1024], *cmd, *service;
	ssize_t bytes;

	bytes = read(fd, buffer, sizeof(buffer) - 1);
	if (bytes == -1)
		return -1;
	if (bytes == 0)
		return 0;
	buffer[bytes] = '\0';
	cmd = buffer;
	service = strchr(buffer, ' ');
	if (service) {
		*service++ = '\0';
		if (strcmp(cmd, "starting") == 0)
			return service_starting(service);
		if (strcmp(cmd, "started") == 0)
			return service_started(service);
		if (strcmp(cmd, "stopping") == 0)
			return service_stopping(service);
		if (strcmp(cmd, "stopped") == 0)
			return service_stopped(service);
		if (strcmp(cmd, "quit") == 0) {
			system("echo quit >/tmp/quit");
			splash_message("Entering", service);
			if (strcmp(service, "shutdown") == 0 ||
			    strcmp(service, "reboot") == 0)
			{
				unlink(SOCKET);
				_exit(0);
			} else
				exit(0);
		}
	}
	errno = EINVAL;
	return -1;
}

static int
make_screen(void)
{
	struct vt_stat vtstat;
	char buffer[1024];
	size_t len;
	WINDOW *cbox;
	FILE *f;
	int null_fd, reset_tio = 0;
	struct termios tio;

	console_fd = open("/dev/console", O_RDWR, 0);
	if (console_fd == -1)
		return -1;
	null_fd = open("/dev/null", O_WRONLY, 0);
	if (null_fd == -1)
		return -1;
	if (ioctl(console_fd, VT_GETSTATE, &vtstat) == -1) {
		close(console_fd);
		return -1;
	}
	if ((splash_fd = open(SPLASH_TTY_PATH, O_RDWR, 0)) == -1) {
		close(console_fd);
		return -1;
	}
	if (ioctl(console_fd, VT_ACTIVATE, SPLASH_TTY) == 0) {
		orig_vt = vtstat.v_active;
		if (orig_vt == SPLASH_TTY)
			orig_vt = 1;
	} else 
		orig_vt = 1;
	if (ioctl(console_fd, VT_WAITACTIVE, SPLASH_TTY) == -1)
		return -1;
	stdout_fd = STDOUT_FILENO;
	stderr_fd = STDERR_FILENO;
	dup2(splash_fd, stdout_fd);
	dup2(splash_fd, stderr_fd);
	dup2(null_fd, STDIN_FILENO);

	mainwnd = initscr();
	if (has_colors()) {
		start_color();
		init_pair(COLOR_BLUE, COLOR_WHITE, COLOR_BLUE);
		init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLUE);
	}
	curs_set(0);
	nodelay(mainwnd, TRUE);
	wrefresh(mainwnd);
	refresh();

	/* We need to set ONLCR to display this properly */
	tcgetattr(splash_fd, &tio);
	if (!(tio.c_oflag & ONLCR)) {
		tio.c_oflag |= ONLCR;
		tcsetattr(splash_fd, TCSANOW, &tio);
		reset_tio = 1;
	}
	if ((f = fopen(LOGO, "r"))) {
		while ((fgets(buffer, sizeof(buffer), f))) {
			len = strlen(buffer);
			write(STDOUT_FILENO, buffer, len);
		}
		fclose(f);
	}
	if (reset_tio) {
		tio.c_oflag &= ~ONLCR;
		tcsetattr(splash_fd, TCSANOW, &tio);
	}

	snprintf(buffer, sizeof(buffer), " OpenRC is %s this host ",
		 starting ? "starting up" : "shutting down");
	cbox = newwin(2, strlen(buffer) + 2, 0, COLS - (strlen(buffer) + 2));
	wcolor_set(cbox, COLOR_BLUE, NULL);
	wborder(cbox, ACS_VLINE, 0, 0, ACS_HLINE, 0, 0, 0, 0);
	mvwprintw(cbox, 0, 1, "%s", buffer);
	wrefresh(cbox);
	mbox = newwin(0, 0, LINES - 4, 0);
	wcolor_set(mbox, COLOR_BLUE, NULL);
	box(mbox, ACS_VLINE, ACS_HLINE);
	wrefresh(mbox);
	splash_progress();
	splash_message("Waiting for", "services");
	return 0;
}

static int
spawn_daemon(const char *runlevel, int opts)
{
	RC_DEPTREE *deptree;
	RC_STRING *s, *t;
	RC_SERVICE state;
	pid_t pid;
	int fd, n;
	size_t i;
	struct sockaddr_un run;
	socklen_t len;

	/* First calculate our deptree */
	deptree = rc_deptree_load();
	if (deptree == NULL)
		return -1;
	svc_todo = rc_deptree_order(deptree, runlevel, opts | RC_DEP_STRICT);
	if (svc_todo == NULL)
		return 0;
	starting = (opts & RC_DEP_START);
	/* Remove things from the todo list we just won't do */
	TAILQ_FOREACH_SAFE(s, svc_todo, entries, t) {
		state = rc_service_state(s->value);
		if ((starting && !(state & RC_SERVICE_STOPPED)) ||
		    (!starting && (state & RC_SERVICE_STOPPED)))
		{
			TAILQ_REMOVE(svc_todo, s, entries);
			free(s);
		}
	}
	TAILQ_INIT(&svc_done);

	unlink(SOCKET);
	if ((fd = make_sock(&len)) == -1 ||
	    bind(fd, (struct sockaddr *)&sun, len) == -1 ||
	    chmod(SOCKET, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP) == -1 ||
	    set_nonblock(fd) == -1 ||
	    listen(fd, NFDS) == -1)
	{
		close(fd);
		fd = -1;
		return -1;
	}

	if ((pid = fork()) != 0) {
		setpgrp(pid, 0);
		close(fd);
		close(console_fd);
		close(splash_fd);
		fd = -1;
		return pid;
	}

	fclose(rc_environ_fd);
	atexit(cleanup);

	make_screen();
	
	for (i = 0; i < NFDS; i++) {
		fds[i].fd = -1;
		fds[i].events = POLLIN;
	}
	fds[0].fd = fd;
	for (;;) {
		for (i = 0; i < NFDS; i++)
			fds[i].revents = 0;
		n = poll(fds, NFDS, -1);
		if (n == -1) {
			if (errno == EAGAIN)
				continue;
			break;
		}
		if (n == 0) /* should not happen, but just in case */
			continue;
		/* Service messages before new connections
		 * This frees existing connections for new ones */
		for (i = 1; i < NFDS; i++) {
			if (fds[i].revents & (POLLIN | POLLHUP)) {
				process_message(fds[i].fd);
				close(fds[i].fd);
				fds[i].fd = -1;
			}
		}
		if (fds[0].revents & (POLLIN | POLLHUP)) {
			for (i = 1; i < NFDS; i++) {
				if (fds[i].fd == -1) {
					len = sizeof(run);
					fds[i].fd = accept(fds[0].fd,
							   (struct sockaddr *)&run,
							   &len);
					set_nonblock(fds[i].fd);
					break;
				}
			}
		}
	}
	return 0;
}

static ssize_t
send_message(const char *message, const char *service)
{
	int fd;
	socklen_t len;
	struct iovec iov[3];
	ssize_t bytes;

	if ((fd = make_sock(&len)) == -1)
		return -1;
	if (connect(fd, (struct sockaddr *)&sun, len) == -1)
		return -1;
	if (service) {
		iov[0].iov_base = UNCONST(message);
		iov[0].iov_len = strlen(message);
		iov[1].iov_base = UNCONST(" ");
		iov[1].iov_len = 1;
		iov[2].iov_base = UNCONST(service);
		iov[2].iov_len = strlen(service);
		bytes = writev(fd, iov, 3);
	} else {
		bytes = write(fd, message, strlen(message));
	}
	close(fd);
	return bytes;
}

int
rc_plugin_hook(RC_HOOK hook, const char *name)
{
	if (rc_yesno(getenv("IN_BACKGROUND")))
		return 0;

	switch (hook) {
	case RC_HOOK_RUNLEVEL_STOP_IN:
		if (strcmp(name, "shutdown") == 0 ||
		    strcmp(name, "reboot") == 0 ||
		    strcmp(name, "single") == 0)
		{
		    	if (rc_yesno(getenv("RC_GOINGDOWN")))
				return spawn_daemon(name, RC_DEP_STOP);
			else
				return spawn_daemon("default", RC_DEP_START);
		}
		break;
	case RC_HOOK_RUNLEVEL_STOP_OUT:
		if (strcmp(name, "shutdown") == 0 ||
		    strcmp(name, "reboot") == 0 ||
		    strcmp(name, "single") == 0)
		    	if (rc_yesno(getenv("RC_GOINGDOWN")))
		    		return send_message("quit", name);
		break;
	case RC_HOOK_RUNLEVEL_START_IN:
		if (strcmp(name, "sysinit") == 0) {
			starting = 1;
			make_screen();
			splash_message("Starting", "sysinit");
		}
		break;
	case RC_HOOK_RUNLEVEL_START_OUT:
		if (strcmp(name, "sysinit") == 0)
			return spawn_daemon("default", RC_DEP_START);
		if (strcmp(name, "boot") != 0)
			return send_message("quit", name);
		break;
	case RC_HOOK_ABORT:
		return send_message("quit", NULL);
	case RC_HOOK_SERVICE_STOP_IN:
		break;
	case RC_HOOK_SERVICE_STOP_NOW:
		return send_message("stopping", name);
	case RC_HOOK_SERVICE_STOP_DONE:
		return send_message("stopped", name);
	case RC_HOOK_SERVICE_STOP_OUT:
		break;
	case RC_HOOK_SERVICE_START_IN:
		break;
	case RC_HOOK_SERVICE_START_NOW:
		return send_message("starting", name);
	case RC_HOOK_SERVICE_START_DONE:
		return send_message("started", name);
	case RC_HOOK_SERVICE_START_OUT:
		break;
	}
	return 0;
}
